#define SEAL_VERSION "2.3.1"
/* #undef SEAL_DEBUG */
#define SEAL_USE_INTRIN
#define SEAL_USE___BUILTIN_CLZLL
#define SEAL_USE___INT128
#define SEAL_USE__ADDCARRY_U64
#define SEAL_USE__SUBBORROW_U64
/* #undef SEAL_USE_MSGSL */
/* #undef SEAL_USE_MSGSL_SPAN */
/* #undef SEAL_USE_MSGSL_MULTISPAN */
